#import "StructureBlend.h"


@implementation StructureBlend : QCPatch

+ (int)executionMode
{
	return 2;
}
+ (BOOL)allowsSubpatches
{
	return NO;
}
+ (int)timeMode
{
	return 1;
}


- (id)initWithIdentifier:(id)fp8
{
	id z=[super initWithIdentifier:fp8];

	if(z)
	{
		// ordinal/normalized
		[inputOrderMode setMaxIndexValue:1];
		// clamp/repeat
		[inputWrapMode setMaxIndexValue:1];
	}

	return z;
}


- (id)setup:(QCOpenGLContext *)context
{


	return context;
}
- (void)cleanup:(QCOpenGLContext *)context
{


}


- (void)enable:(QCOpenGLContext *)context
{


}
- (void)disable:(QCOpenGLContext *)context
{


}


- (BOOL)execute:(QCOpenGLContext *)context time:(double)time arguments:(NSDictionary *)arguments
{
	QCStructure *structure = [inputStruct structureValue];
	double index = [inputIndex doubleValue];

	if(structure == nil || [structure count] == 0)	// if there isn't a buffer, or it's empty; there's no point evaluating more.
		return YES;
	// Note: index is a floating point value.  Never check index with equality only, it must have greater-than or less-than to be useful.
	switch([inputOrderMode indexValue])
	{
		default:
		case 0:	// ordinal
			switch([inputWrapMode indexValue])
			{
				default:
				case 0:	// clamp
					if(index <= 0.0)
						[outputVirtual setRawValue:[structure memberAtIndex:0]];
					else if(index >= [structure count]-1)
						[outputVirtual setRawValue:[structure memberAtIndex:[structure count]-1]];
					else
						[outputVirtual setRawValue:[structure memberAtIndex:(int)index]];
					break;
				case 1:	// repeat
					// special case: if there's only 1 element in the array, our fmod will try to 
					// divide by zero.  repeat is trivial in count==1 case anyway :)
					if([structure count] == 1)
						[outputVirtual setRawValue:[structure memberAtIndex:0]];
					else
					{
						index = fmod(index,[structure count] - 1.0);
						if(index < 0.0)	// -3%5 is -3, not 2. -7%5 is -2, not 3 
							index += (double)[structure count];
						[outputVirtual setRawValue:[structure memberAtIndex:(int)index]];
					}
					break;
			}
			break;
		case 1:	// normalized
			switch([inputWrapMode indexValue])
			{
				default:
				case 0:	// clamp
					if(index <= 0.0)
						[outputVirtual setRawValue:[structure memberAtIndex:0]];
					else if(index >= 1.0)
						[outputVirtual setRawValue:[structure memberAtIndex:[structure count]-1]];
					else 
						[outputVirtual setRawValue:[structure memberAtIndex:
										(int) (index * ((double)[structure count] - 1.0)) 
						]];
					break;
				case 1:	// repeat
					// special case: if there's only 1 element in the array, our fmod will try to 
					// divide by zero.  repeat is trivial in count==1 case anyway :)
					if([structure count] == 1)
						[outputVirtual setRawValue:[structure memberAtIndex:0]];
					else
					{
						if(index > 1.0 || index < 0.0)
							index = fmod(index, 1.0);
						if(index < 0.0)	// -3%5 is -3, not 2. -7%5 is -2, not 3 
							index += 1.0;
						[outputVirtual setRawValue:[structure memberAtIndex:
										(int) (index * ((double)[structure count] - 1.0)) 
						]];
					}
					break;
			}
			break;
	}

	return YES;
}

@end
