#import <QuartzCore/CIImageAccumulator.h>
#import "StructureRecordAdv.h"
#import "QCGLCGImage.h"
#import "QCGLCIImage.h"
#import "QCGLBitmapImage.h"

@implementation StructureRecordAdv : QCPatch

@class QCGLCIImage;

+ (int)executionMode
{
	return 2;
}
+ (BOOL)allowsSubpatches
{
	return NO;
}
+ (int)timeMode
{
	return 1;
}


- (id)initWithIdentifier:(id)fp8
{
	id z=[super initWithIdentifier:fp8];

	if(z)
	{
		// inputMode indexValues:
		//	0 = use gate raw
		//	1 = "buffer at once" mode (now called Record To Buffer Size)
		//	2 = DSP mode (0th index is always the current sample, data slides down the array)
		//	3 = record on rising edge (one-shot mode)
		//	4 = stop when full mode
		[inputMode setMaxIndexValue:4];
		[inputRateDivider setIndexValue:1];
		[inputMaxSamples setIndexValue:10];	// don't start in unlimited mode by default :)
		index = 0;
		buffer = nil;
		//imageAccumulators = [[NSMutableDictionary alloc] initWithCapacity:[inputMaxSamples indexValue]];
	}

	return z;
}


- (id)setup:(QCOpenGLContext *)context
{
	return context;
}

- (void)cleanup:(QCOpenGLContext *)context
{
	if(buffer != nil)
	{
		// since we copy all incoming items, we need to release them
		[buffer removeAllObjects];
		[buffer release];
		//[imageAccumulators release];
		buffer = nil;
	}
}


- (void)enable:(QCOpenGLContext *)context
{


}
- (void)disable:(QCOpenGLContext *)context
{


}

/* this helper function clones objects that don't adhere to the NSCopying protocol correctly, intelligently, or at all 
	Currently, we handle: 
	   * QCGLBitmapImages (we might want to use the parent class QCImage?) - copy is too expensive.
	   * QCStructure (NSCopying doesn't work on this class)
 */
id copyItem(id object)
{
	id newObject;
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	if( [object class] == [QCGLBitmapImage class])
	{
		CGRect r;
		r.origin.x=0.0;
		r.origin.y=0.0;
		r.size.width=CGImageGetWidth([object CGImage]);
		r.size.height=CGImageGetHeight([object CGImage]);
		
		CIImageAccumulator *cia = [[CIImageAccumulator alloc] initWithExtent:r format:kCIFormatARGB8];
		
		if(cia == nil)
			return nil;
			
		[cia setImage:[object CIImage]];

		newObject = cia;
	}
	else if ([object class] == [QCStructure class])
	{
		// This is kind of a hack.  Structures store their key/value pairs in 2 GFLists.  We simply reuse these lists
		// to perform a "shallow copy (meaning we only clone pointers, not data).  
		GFList *list = [object _list];
		QCStructure *qcs = [[QCStructure alloc] 
			initWithMembers:(id *)[list _values]
			forKeys:(id *)[list _keys]
			count:[object count]];
		
		newObject = qcs;
	}
	else
		newObject = [object copy];
		
	[pool release];
	
	return newObject;
}

- (BOOL)execute:(QCOpenGLContext *)context time:(double)time arguments:(NSDictionary *)arguments
{
	if([inputRateDivider indexValue] <= 0)	// can't have a negative no zero rate...
		[inputRateDivider setIndexValue:1];
		
	// clear on clear or on gate rising edge if clear-on-gate is set
	if([inputClear booleanValue] == TRUE || [inputClearWithObject booleanValue] == TRUE ||
		([inputClearOnGate booleanValue] == TRUE && oldGateValue == FALSE && [inputGate booleanValue] == TRUE))
	{
		//NSLog(@"Clearing array...\n");
		index = 0;
		if(buffer != nil)
		{
			// empty the buffer!
			[buffer removeAllObjects];
			[buffer release];
			buffer = nil;
			divider = 0;
		}
		if([inputClearWithObject booleanValue] == TRUE)
		{
			buffer = [[NSMutableArray alloc] initWithCapacity:1+[inputMaxSamples indexValue]];
			while([buffer count] < [inputMaxSamples indexValue])
			{
				id newItem = copyItem([inputVirtual rawValue]);
				[buffer addObject:newItem];
				[newItem release];
			}
		}
		else
			[outputStruct setStructureValue:nil];
	}
	
	if([inputGate booleanValue] == TRUE || (index != 0 && [inputMode indexValue] == 1))
	{
		//bNSLog(@"Recording...");
		if(buffer == nil)
			buffer = [[NSMutableArray alloc] initWithCapacity:1+[inputMaxSamples indexValue]];
		if(divider % [inputRateDivider indexValue] == 0 && [inputVirtual rawValue])
		{
			if([inputMaxSamples indexValue] > 0)	// finite buffer mode
			{
				if([inputMode indexValue] < 2)
				{
					// raw gate and buffer-at-once modes
					// growing buffer mode
					if([inputMaxSamples indexValue] > [buffer count])
					{
						id newItem = copyItem([inputVirtual rawValue]);
						[buffer addObject:newItem];
						[newItem release];
					}
					// replace buffer mode (common case)
					else
					{
						//NSLog(@"Replace mode...\n");
						id newItem = copyItem([inputVirtual rawValue]);
						[buffer replaceObjectAtIndex:index
							withObject:newItem];
						[newItem release];
					}
				}
				else if([inputMode indexValue] == 2)	// dsp mode
				{
					id newItem = copyItem([inputVirtual rawValue]);
					[buffer insertObject:newItem atIndex:0];
					[newItem release];
				}
				else if([inputMode indexValue] == 3 && oldGateValue == FALSE && [inputGate booleanValue] == TRUE)	// one-shot
				{
					// growing buffer mode
					if([inputMaxSamples indexValue] > [buffer count])
					{
						//NSLog(@"Growing mode... (%i) %i vs %i\n",video,[buffer count],[inputMaxSamples indexValue]);
						id newItem = copyItem([inputVirtual rawValue]);
						[buffer addObject:newItem];
						[newItem release];
					}
					// replace buffer mode (common case)
					else
					{
						//NSLog(@"Replace mode...\n");
						id newItem = copyItem([inputVirtual rawValue]);
						[buffer replaceObjectAtIndex: index
							withObject: newItem];
						[newItem release];
					}
				}
				else if([inputMode indexValue] == 4 && index < [inputMaxSamples indexValue])
				{
					id newItem = copyItem([inputVirtual rawValue]);
					[buffer addObject:newItem];
					[newItem release];
				}
				
				// when the user changes the buffer size on the fly, we can have extra data at the end if they make it smaller.
				while([buffer count] > [inputMaxSamples indexValue])
					[buffer removeLastObject];
				
				[outputIndex setIndexValue:index];
				
				// index changes in a few different ways depending on mode.
				if([inputMode indexValue] == 4)
				{
					// index+1 so that we never set index to outside the buffer
					if(index+1 < [inputMaxSamples indexValue])
						index++;
				}
				else if([inputMode indexValue] == 3)
				{
					if( oldGateValue == FALSE && [inputGate booleanValue] == TRUE)
						index = (index+1)%[inputMaxSamples indexValue];
				}
				else if([inputMode indexValue] == 2)
					index = 0;
				else
					index = (index+1)%[inputMaxSamples indexValue];
				
			}
			else	// unlimited buffer mode.  Just keep appending
			{
				id newItem = copyItem([inputVirtual rawValue]);
				[buffer addObject:newItem];
				[outputIndex setIndexValue:[buffer count]];
				[newItem release];
			}
			
			/* process buffer: can't use the CIImageAccumulators as-is, need to drop them to [cia image] for the structure */
			NSMutableArray *processed = [[NSMutableArray alloc] initWithArray:buffer];
			unsigned int i;
			for(i=0;i<[processed count];++i)
				if( [[processed objectAtIndex:i] class] == [CIImageAccumulator class] )
				{
					QCGLCIImage *img = [[QCGLCIImage alloc] initWithCIImage:[[processed objectAtIndex:i] image] options:nil];
					[processed replaceObjectAtIndex:i 
						withObject:img];
					[img release];
				}
			
			structure = [[QCStructure alloc] initWithArray:processed];
			[processed release];
			
			[outputStruct setStructureValue: structure];
			[structure release];
		}
		++divider;
	}
	
	oldGateValue = [inputGate booleanValue];
	
	return YES;
}

@end
